; ------------------------------------------------------------------------
; Tetris game in AutoLISP (C) 2004, written by Roman S. Klyujkov
; Contact e-mail: rsk_comp <at> mail.ru
; Worked on AutoCAD R12 only, control keys like in DOS Navigator's tetris
; Commands to load & start the game:
; 	(load "tet")
; 	tet
; If you want to start with specified level, use command:
; 	(setq tet:startlevel <number [1..10]>)
; and then start the game with tet command
; Anti-Boss key: <Ins> ;)
; ------------------------------------------------------------------------

(defun tet:rect (p1 p2 / p3 p4)
	(setq p3 (list (car p2) (cadr p1)) p4 (list (car p1) (cadr p2)))
	(grdraw p1 p3 255)
	(grdraw p3 p2 255)
	;(grdraw p2 p4 255)
	(grdraw p4 p1 255)
)

(defun tet:wide (p1 p2 w)
	(list
		(list (- (car p1) w) (- (cadr p1) w))
		(list (+ (car p2) w) (+ (cadr p2) w))
	)
)

(defun *error* (msg)
	(if stackvar (while (popvar) nil))
	(redraw)
	(princ msg)
	(princ "\n")
	nil
)

(setq tet:fig '(
	; xx red
	; xx
	(1
		((0 0) (1 0) (0 1) (1 1))
	)
	;  xx green
	; xx
	(3
		((-1 0) (0 0) (0 1) (1 1))
		((0 -1) (0 0) (-1 0) (-1 1))
	)
	; x  cyan
	; xx
	; x
	(4
		((0 -1) (0 0) (0 1) (1 0))
		((-1 0) (0 0) (0 1) (1 0))
		((-1 0) (0 -1) (0 0) (0 1))
		((-1 0) (0 -1) (0 0) (1 0))
	)
	; x  yellow
	; x
	; xx
	(2
		((0 1) (0 0) (0 -1) (1 -1))
		((-1 -1) (0 -1) (1 -1) (1 0))
		((0 0) (1 0) (1 -1) (1 -2))
		((0 -1) (0 0) (1 0) (2 0))
	)
	;  x blue
	; xx
	; x
	(5
		((0 -1) (0 0) (1 0) (1 1))
		((-1 1) (0 1) (0 0) (1 0))
	)
	;  x white
	;  x
	; xx
	(7
		((0 -1) (1 -1) (1 0) (1 1))
		((-1 0) (0 0) (1 0) (1 -1))
		((0 -2) (0 -1) (0 0) (1 0))
		((0 0) (0 -1) (1 -1) (2 -1))
	)
	; x  magenta
	; x
	; x
	; x
	(6
		((0 -2) (0 -1) (0 0) (0 1))
		((-2 0) (-1 0) (0 0) (1 0))
	)
))

(setq tet:template '(
	; 譨 
	(0 0) (0.9 0)
	(0.9 0) (0.9 0.9)
	(0.9 0.9) (0 0.9)
	(0 0.9) (0 0)
	; ७ 
	;(0.1 0.1) (0.8 0.1)
	(0.8 0.1) (0.8 0.8)
	(0.8 0.8) (0.1 0.8)
	;(0.1 0.8) (0.1 0.1)
))

(defun tet:revolvfig (coos)
	(mapcar
		'(lambda (coo)
			(list (- (cadr coo)) (car coo))
		)
		coos
	)
)

(defun tet:bbox (lst / p x0 y0 x1 y1)
	(setq
		p (car lst)
		x0 (car p) y0 (cadr p)
		x1 x0 y1 y0
	)
	(foreach
		p
		(cdr lst)
		(setq
			x0 (min x0 (car p))
			y0 (min y0 (cadr p))
			x1 (max x1 (car p))
			y1 (max y1 (cadr p))
		)
	)
	(list x0 y0 x1 y1)
)

(defun tet:nthreplace (n lst func)
	(mapcar
		'(lambda (xxx)
			(if (zerop n)
				(setq xxx (apply func (list xxx)))
			)
			(setq n (1- n))
			xxx
		)
		lst
	)
)

; 頥: (.Y ᪠1 ... ᪠N), ᫨ 室   ࠭: nil
(defun tet:bitmaskfig (coos x y / bbox ymin n mask coo)
	; 塞 . אַ.: (xmin ymin xmax ymax)
	(setq bbox (tet:bbox coos))
	; ஢塞  室   ࠭
	(if (minusp (+ (car bbox) x))
		nil
		(progn
			; 室 - ப  ⬠᪥: 1+ymax-ymin
			(setq n (1+ (- (cadddr bbox) (setq ymin (cadr bbox)))))
			; . ﬨ
			(repeat n (setq mask (cons 0 mask)))
			; ᨬ  न  
			(foreach coo coos
				(setq n (- (cadr coo) ymin))
				(setq mask
					(tet:nthreplace n mask
						'(lambda (old)
							(logior old (lsh 1 (+ (car coo) x)))
						)
					)
				)
			)
			(cons (+ y ymin) mask)
		)
	)
)

(defun tet:shleft (bmfig)
	(cons
		(car bmfig)
		(mapcar
			'(lambda (x)
				(if (= (logand x 1) 0)
					(lsh x -1)
					nil
				)
			)
			(cdr bmfig)
		)
	)
)

; maxcol=(1- (car txy))
(defun tet:shright (bmfig maxcol)
	(setq maxcol (lsh 1 maxcol))
	(cons
		(car bmfig)
		(mapcar
			'(lambda (x)
				(if (zerop (logand x maxcol))
					(lsh x 1)
				)
			)
			(cdr bmfig)
		)
	)
)

(defun tet:collisionbound (bmfig / x pred)
	(setq pred T)
	(foreach
		x
		(cdr bmfig)
		(if (not x) (setq pred nil))
	)
	pred
)

(defun tet:overflowright (bmfig maxcol / flag)
	(setq maxcol (lsh 1 maxcol) flag T)
	(foreach
		x
		bmfig
		(if (= (logand x maxcol) maxcol)
			(setq flag nil)
		)
	)
	flag
)

(defun tet:collisionbm (bmfig bmglass / y flag bm)
	(setq y (car bmfig) flag T)
	(foreach
		bm
		(cdr bmfig)
		(if (/= (logand bm (nth y bmglass)) 0)
			(setq flag nil)
		)
		(setq y (1+ y))
	)
	flag
)

(defun tet:mergebm (bmfig bmglass / n)
	(setq n (car bmfig) bmfig (cdr bmfig))
	(mapcar
		'(lambda (x / carbmfig)
			(if (and (= n 0) (car bmfig))
				(progn
					(setq carbmfig (car bmfig) bmfig (cdr bmfig))
					(logior x carbmfig)
				)
				(progn
					(setq n (1- n))
					x
				)
			)
		)
		bmglass
	)
)

(defun tet:dropdownbm (bmfig bmglass / dy)
	(setq dy -1)
	(while (and (>= (car bmfig) 0) (tet:collisionbm bmfig bmglass))
		(setq
			bmfig (cons (1- (car bmfig)) (cdr bmfig))
			dy (1+ dy)
		)
	)
)

(defun tet:drawfig (coos)
	(apply 'append
		(mapcar
			'(lambda (coo)
				(mapcar
					'(lambda (xy)
						(mapcar
							'+
							coo
							xy
						)
					)
					tet:template
				)
			)
			coos
		)
	)
)

(defun tet:setmatr (p0 dt x y)
	(list
		(list dt 0 0 (+ (car p0) (* x dt)))
		(list 0 dt 0 (+ (cadr p0) (* y dt)))
		(list 0 0 1 0)
		(list 0 0 0 1)
	)
)

(defun tet:glass (p0 txy dt bmglass / pp rc x y n)
	(grclear)
	(setq pp (list (+ (car p0) (* (car txy) dt)) (+ (cadr p0) (* (cadr txy) dt))))
	(setq rc (tet:wide p0 pp (/ dt 4.0)))
	(tet:rect (car rc) (cadr rc))
	(setq rc (tet:wide p0 pp (/ dt 2.0)))
	(tet:rect (car rc) (cadr rc))
	(setq y 0)
	(repeat (cadr txy)
		(setq
			n (car bmglass)
			x 0
		)
		(while (> n 0)
			(if (= (logand n 1) 1)
				(grvecs (cons 7 tet:template) (tet:setmatr p0 dt x y))
			)
			(setq
				n (lsh n -1)
				x (1+ x)
			)
		)
		(setq
			bmglass (cdr bmglass)
			y (1+ y)
		)
	)
)

(defun tet:removefullrow (bmglass fullrow / n bmglassnew row)
	(setq n 0)
	(foreach
		row
		bmglass
		(if (= row fullrow)
			(setq n (1+ n))
			(setq bmglassnew (cons row bmglassnew))
		)
	)
	(repeat n (setq bmglassnew (cons 0 bmglassnew)))
	(cons n (reverse bmglassnew))
)

(defun tet:leveldelay (level / delays)
	; delay by level in msecs
	(setq delays
		'(0 1600 1200 1000 600 500 400 300 200 100 80)
	)
	(/ (nth level delays) 1000.0 86400.0) ; 86400.0
)

(setq tet:startlevel 4)

(defun c:tet (/ vs vc txy dt p0 n coos matr fig fullrow refresh x y rd key coosnew matrnew bmfig bmfignew bmglass endofturn color tim0 delays delay level lines endofgame temp tempnew nfig rot str)
	(setq
		txy '(12 19)
		fullrow (1- (lsh 1 (car txy)))
		vs (getvar "VIEWSIZE")
		vc (getvar "VIEWCTR")
		dt (/ vs (+ (cadr txy) 2))
		p0 (list (- (car vc) (/ (* (car txy) dt) 2.0)) (- (cadr vc) (/ (* (cadr txy) dt) 2.0)))
		key 0 x 3 y 3 color 3
		level tet:startlevel
		delay (tet:leveldelay level)
		lines 0
		endofgame nil
	)
	; . ﬨ ⬠ ⠪
	(repeat (+ (cadr txy) 5) (setq bmglass (cons 0 bmglass)))
	; 㥬 ⠪
	(tet:glass p0 txy dt bmglass)
	(while (/= key 27)
		(setq
			fig (nth (rem (fix (* (getvar "TDUSRTIMER") 555 86400.0)) 7) tet:fig)
			color (car fig)
			rot (cdr fig)
			x (/ (car txy) 2) y (1- (cadr txy))
			;|color (nth nfig tet:color)
			rot (nth nfig tet:rot)|;
			coosnew (car rot);|(cdr fig)|; matrnew (tet:setmatr p0 dt x y) bmfignew (tet:bitmaskfig coosnew x y)
			rot (cdr rot)
			tempnew (tet:drawfig coos)
			coos coosnew matr matrnew bmfig bmfignew temp tempnew
			tim0 (getvar "TDUSRTIMER")
			endofturn nil
			refresh T
		)
		(if (not (tet:collisionbm bmfig bmglass))
			(setq
				endofturn T
				key 27
				endofgame T
			)
		)
		(while (not endofturn)
			(if refresh
				(progn
					(setq tempnew (tet:drawfig coosnew))
					(grvecs (cons 0 temp) matr)
					(setq coos coosnew matr matrnew bmfig bmfignew temp tempnew refresh nil)
					(grvecs (cons color temp) matr)
				)
			)
			(while (not refresh)
				(if (>= (- (getvar "TDUSRTIMER") tim0) delay)
					(progn
						(setq
							y (1- y)
							refresh T
							; ᬥ頥 ⬠ 
							bmfignew (cons (1- (car bmfig)) (cdr bmfig))
							matrnew (tet:setmatr p0 dt x y)
							coosnew coos
							tim0 (getvar "TDUSRTIMER")
						)
						(if (or (= (car bmfig) 0) (not (tet:collisionbm bmfignew bmglass)))
							(progn
								(setq bmglass (tet:mergebm bmfig bmglass))
								(grvecs (cons 7 temp) matr)
								;
								(setq
									bmglass (tet:removefullrow bmglass fullrow)
									n (car bmglass)
									bmglass (cdr bmglass)
								)
								;
								(if (> n 0)
									(progn
										(tet:glass p0 txy dt bmglass)
										(setq lines (+ lines n))
										(if (> lines (- 90 (* (- 9 tet:startlevel) 10)))
											(setq
												level (+ 1 (/ (fix (- lines (- 90 (* (- 9 tet:startlevel) 10)))) 20) tet:startlevel)
												level (min level 10)
												level (max level tet:startlevel)
												delay (tet:leveldelay level)
											)
										)
										(gc)
										(princ "\rLines: ") (princ lines)
										(princ " Level: ") (princ level)
										(princ " Delay: ") (princ (fix (* delay 86400.0 1000.0)))
									)
								)
								(setq endofturn T)
							)
						)
					)
				)
				(if (and (not refresh) (= (car (setq rd (grread T 6 1))) 2))
					(progn
						(setq key (cadr rd))
						;(princ rd)
						(cond
							; boss incoming!
							((= key 210)
								(redraw)
								(setq str "")
								(while (not refresh)
									(princ (strcat "\r" ": " str " "))
									(setq rd (grread) key (cadr rd))
									(if (= (car rd) 2)
										(cond
											((= key 27) (setq refresh T))
											((= key 13)
												(command str)
												(princ "\n")
												(setq str "")
											)
											((= key 8)
												(if (/= str "")
													(setq str (substr str 1 (1- (strlen str))))
												)
											)
											(T
												(setq tet:f (open "$$$" "w")) (write-char key tet:f) (close tet:f)
												(setq tet:f (open "$$$" "r")) (setq str (strcat str (read-line tet:f))) (close tet:f)
											)
										)
										;|(progn
											(write-char (cadr rd))
											(if (= (cadr rd) 27)
												(setq refresh T)
											)
										)|;
									)
								)
								(setq refresh nil)
								(tet:glass p0 txy dt bmglass)
							)
							; 
							((= key 200)
								(if (null rot) (setq rot (cdr fig)))
								(setq
									coosnew (car rot)
									bmfignew (tet:bitmaskfig coosnew x y)
								)
								(if (and bmfignew (tet:overflowright bmfignew (car txy)) (tet:collisionbm bmfignew bmglass))
									(setq refresh T rot (cdr rot))
								)
							)
							; 
							((= key 203)
								(setq
									x (1- x)
									matrnew (tet:setmatr p0 dt x y)
									coosnew coos
									bmfignew (tet:shleft bmfig)
								)
								(if (and (tet:collisionbound bmfignew) (tet:collisionbm bmfignew bmglass))
									(setq refresh T)
									(setq x (1+ x))
								)
							)
							; ࠢ
							((= key 205)
								(setq
									x (1+ x)
									matrnew (tet:setmatr p0 dt x y)
									coosnew coos
									bmfignew (tet:shright bmfig (1- (car txy)))
								)
								(if (and (tet:collisionbound bmfignew) (tet:collisionbm bmfignew bmglass))
									(setq refresh T)
									(setq x (1- x))
								)
							)
							; 
							((= key 208)
								(setq
									n (tet:dropdownbm bmfig bmglass)
									y (- y n)
									bmfignew (cons (- (car bmfig) n) (cdr bmfig))
									matrnew (tet:setmatr p0 dt x y)
									coosnew coos
									refresh T
								)
							)
							; esc
							((= key 27) (setq refresh T endofturn T))
						)
					)
				)
			)
		)
	)
	(if endofgame
		(progn
			(princ "\nScore: ")
		)
		(princ "\r                                \r")
	)
	(redraw)
)